#include "tdbd.h"

#define MODE_16CH 0
#define ICH_HEADER_SIZE 16
#define Sleep(x) usleep(1000*x)
#define SEC_PER_PACKET1 4e-3
#define SEC_PER_PACKET2 4.35e-3
#define DA_BS_TO_PS(x) ((x)/8)
#define AD_BS_TO_PS(x) ((x)/32)
#define MAX_FRAMES_PER_PACKET 192
#define AD_CHANNELS 16
#define DA_CHANNELS 4
#define TDBD_FORAMT WIOS_PCM_FORMAT_S16
#define NOT_USE (1<<29)

#define GAIN_MIN 0
#define GAIN_MAX 96

typedef struct ich_io_t ICH_IO_T;

typedef struct {
	ICH_IO_T  * handle;
	wios_buffer_t tmp_buf;
	UCHAR     * buffer;
	UCHAR     * bp;
	int       msize;
	unsigned long header[4];
	int samp_au;
} tdbd_priv_t;

static int common_samp_au;
static int common_loops;
static int common_sec;
static int capture_req_size;

static uchar device_name[64] = "/dev/sinichusb0";

char output_wav_file_name[128];
char output_ich_file_name[128];

static unsigned char	audio_board_num_table[] = { 0x5A, 0xA5, 0x4B, 0xB4 };

static unsigned long	audio_outpkt_tbl[7]={
	1536,	// 48k
	1024,	// 32k
	512,	// 16k
	256,	// 8k
	1536,	// 44.1k
	768,	// 22.05k
	384		// 11.025k
};

static unsigned long	audio_inpkt_tbl[7]={
	6144,	// 48k
	4096,	// 32k
	2048,	// 16k
	1024,	// 8k
	6144,	// 44k
	3072,	// 22k
	1536	// 11k
};

static int tdbd_supported_rate[] = {
	48000,
	32000,
	16000,
	8000,
	44100,
	22000,
	11000
};



/* STATIC FUNCTION PROTOTYPE */
/* -------------------------------------------------- */

static char * check_file_ext(const char * name);
static int check_sampling_rate(int rate);
static int check_ad_rate(wios_file_t * ftad);
static int check_gain(int gain);
static int check_output_file_name(const char * name);
static int check_da_format(wios_file_t * ft);

static int priv_init(wios_file_t * ft);
static void tdbd_write(wios_file_t * ft);
static void tdbd_read(wios_file_t * ft);

void output_16ch_wav(wios_file_t * ft);

static int codec_control_set(uchar msg_flag, struct ich_io_t *io,int set_gain_val);
static int DisplayDescriptors(HANDLE hDev);
static int DisplayAudioSampleRate(uchar msg_flag, uchar fpga_rdata);

static int hex_check(char *p);
static int ahextoi(char hex);
static int ahextoi2(char *hex_char, int digit);



/* HANDLER FUNCTION */
/* -------------------------------------------------- */

static int device_init(wios_file_t * ft)
{
	int ret;
	unsigned char sample_rate_ch_set = 0;
	uchar fpga_rdata;
	tdbd_priv_t * priv = ft->priv;
	wios_signal_t * sig = &ft->signal;
	static ICH_IO_T * dp = NULL;

	/* 同期モードの際は初期化は1回のみ */
	if (dp) {
		ft->dp = dp;
		return WIOS_SUCCESS;
	}
	
	if ((ret = sinich_sys_open(&dp, (uchar *)ft->dev_name)) != 0)
		WIOS_ERROR_S("driver open error");

	sample_rate_ch_set |= priv->samp_au;
	sample_rate_ch_set |= (MODE_16CH << 3);

	if (sinich_pcm_ctl(dp, sample_rate_ch_set) != 0)
		WIOS_ERROR_S("cannot set sampling status");

	if (sinich_sys_get_reg(dp, 0x01, &fpga_rdata) != 0)
		WIOS_ERROR_S("cannot set sampling status");

	DisplayAudioSampleRate(print_codec_flag, fpga_rdata);

	if (sinich_sys_set_reg(dp, 0x0B, 0x01) != 0)
		WIOS_ERROR_S("registar set error");
	
	if (codec_control_set(print_codec_flag, dp, ft->gain) < 0)
		WIOS_ERROR_S("cannot set codec");

	/* EP4がスタートしていたら止める */
	if (sinich_sys_get_reg(dp, 0x04, &fpga_rdata) != 0)
		WIOS_ERROR_S("reg read error");
	if (fpga_rdata)
		if (sinich_sys_set_reg(dp, 0x04, 0x00) != 0)
			WIOS_ERROR_S("set audio in enable registar failed");
	
	/* EP2がスタートしていたら止める */
	if (sinich_sys_get_reg(dp, 0x03, &fpga_rdata) != 0)
		WIOS_ERROR_S("reg read error");
	if (fpga_rdata)
		if (sinich_sys_set_reg(dp, 0x03, 0x00) != 0)
			WIOS_ERROR_S("set audio in enable registar failed");
	
	if (ft->io_type == wios_da) {
		/* 最初の1回分の再生データを書きこんでおく */
		ret = sinich_pcm_write(dp, priv->buffer, sig->buf_size*4);
		if (ret < 0)
			WIOS_ERROR("pcm write size error %d", ret);
		priv->bp += ret;
		ft->success_byte_cnt += ret;
		sig->loops -= (ret/sig->buf_size);
	}

	if(print_descr_flag) {
		/* USBディスクリプタを表示 */
		ret = DisplayDescriptors(dp->dev);
		if (ret == -1)
			WIOS_ERROR_S("GET USB device descriptor error!\n");
	}

	ft->dp = dp;

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static wios_file_t * open_read_wav(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	int i,j;
	int sec=info->sec;
	int frames;
	long long done = 0;
	wios_file_t * ft;
	tdbd_priv_t * priv;
	wios_signal_t * sig;
	wios_fmtinfo_t * fmt;
	wios_execinfo_t * exec;
	wios_buffer_t * tmp_bp;

	/* WAVファイルを開いてフォーマットチェック */
	if ((ft = open_read_(path, dev, info, priv_init)) == NULL)
		goto error;
	if (check_da_format(ft) == WIOS_EOF)
		goto error;

	priv = ft->priv;
	sig  = &ft->signal;
	fmt  = &ft->fmtinfo;
	exec = &ft->signal.execinfo;

	priv->samp_au = check_sampling_rate(fmt->rate);
	sig->buf_size = audio_outpkt_tbl[priv->samp_au];
	sig->byte_per_item = byte_per_item[fmt->format];
	exec->periodsize = DA_BS_TO_PS(sig->buf_size);
	
	tmp_bp = &priv->tmp_buf;
	if (wios_buffer_malloc(ft, tmp_bp, sig->all_frames) == NULL)
		goto error;

	/* 一時バッファにPCMデータを全て読み込む */
	frames = lsf_file_read(ft, tmp_bp->fl, sig->all_frames);
	if (frames != sig->all_frames)
		WIOS_ERROR("cannot read all wav frames. only read %d frames", frames);

	priv->msize = sizeof(int16_t)*DA_CHANNELS*sig->all_frames;

	if ((priv->buffer = priv->bp = (UCHAR *)malloc(sizeof(int16_t)*DA_CHANNELS*sig->all_frames)) == NULL)
		WIOS_ERROR_S("write buffer allocate error");
	
	/* PCMデータの並びをWAV配列からoch配列に変える */
	while (done < sig->all_frames) {
		int16_t * bp=(int16_t *)priv->buffer;
		int nsamp = ft->signal.execinfo.periodsize/2;

		for (i = 0; i < nsamp; i++) {
			for (j = 0; j < DA_CHANNELS; j++) {
				if (done*DA_CHANNELS+i*DA_CHANNELS+j >= sig->all_frames*DA_CHANNELS)
					break;
				bp[done*DA_CHANNELS+j*nsamp+i] = 
					WIOS_CONVERT_32_TO_16BIT(tmp_bp->fl[done*DA_CHANNELS+i*DA_CHANNELS+j]);
			}
		}
		done += nsamp;
	}

	/* WAVファイルはもう使わないので閉じる */
	if (ft->fp) {
		sf_close(ft->fp);
		ft->fp = NULL;
	}

	/* loopsとcapture_req_sizeの値を指定 */
	if (sec != WIOS_ALL_SEC) {
		double sec_per_packet;
		sec_per_packet = ((priv->samp_au < 4) ? SEC_PER_PACKET1 : SEC_PER_PACKET2);
		sig->loops = (int)(sec/sec_per_packet);
		sig->loops++;
		capture_req_size = sig->loops*sig->buf_size*AD_CHANNELS/DA_CHANNELS;
	} else {
		sig->loops = sig->loops_org = priv->msize/sig->buf_size+1;
		capture_req_size = priv->msize*AD_CHANNELS/DA_CHANNELS;
	}

	/* 同期モードの時デバイスの初期化の際に必要 */
	ft->gain = set_gain_val;

	/* 録音フォーマットを揃える */
	common_samp_au = priv->samp_au;
	common_loops   = sig->loops;
	common_sec     = info->sec;

	return ft;

error:
	return NULL;
}

static wios_file_t * open_read_och(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	int sec=info->sec;
	int fsize = 0;
 	uint fpos_start, fpos_end;
	wios_file_t * ft;
	tdbd_priv_t * priv;
	wios_signal_t * sig;
	wios_fmtinfo_t * fmt;
	wios_execinfo_t * exec;

	/* ochファイルを開いて構造体初期化 */
	if ((ft = malloc(sizeof(wios_file_t))) == NULL) 
		WIOS_ERROR_S("cannot allocate file structure");
	if (priv_init(ft) == WIOS_EOF)
		WIOS_ERROR_S("cannot allocate file structure");

	ft->io_type = wios_da;
	ft->file_name = path ? path : TDBD_DEFAULT_OCHFILE_NAME;
	ft->dev_name  = dev  ? dev  : TDBD_DEFAULT_DEVICE_NAME;
	if (info) ft->signal.execinfo = *info;
	else wios_execinfo_init(&ft->signal.execinfo);

	priv = ft->priv;
	sig  = &ft->signal;
	fmt  = &ft->fmtinfo;
	exec = &ft->signal.execinfo;

	if ((ft->fp = fopen(ft->file_name, "rb")) == NULL)
		WIOS_ERROR_S("och file open error");

	if (fread(priv->header, sizeof(uchar), 0x10, ft->fp) != 0x10)
		WIOS_ERROR_S("file header read error");

	priv->samp_au = priv->header[1];
	sig->buf_size = audio_outpkt_tbl[priv->samp_au];
	sig->byte_per_item = byte_per_item[WIOS_PCM_FORMAT_S16];
	exec->periodsize = DA_BS_TO_PS(sig->buf_size);

	fmt->rate = tdbd_supported_rate[priv->samp_au];
	fmt->channels = DA_CHANNELS;
	fmt->format = WIOS_PCM_FORMAT_S16;

	/* ochファイルサイズを取得 */
	fseek(ft->fp, 0, SEEK_SET);
	fpos_start = ftell(ft->fp);
	fseek(ft->fp, 0, SEEK_END);
	fpos_end = ftell(ft->fp);
	fsize=(int)(fpos_end - fpos_start);
	fseek(ft->fp, 0x10 , SEEK_SET);
	fsize-=4;
	priv->msize=(fsize/sig->buf_size)*sig->buf_size;

	if ((priv->buffer = priv->bp = (UCHAR *)malloc(priv->msize)) == NULL)
		WIOS_ERROR_S("write buffer allocate error");

	/* PCMデータを先にすべて読み込んでおく */
	if (fread(priv->buffer, sizeof(uchar), priv->msize, ft->fp) == 0)
		WIOS_ERROR_S("File read error");

	/* loopsとcapture_req_sizeの値を指定 */
	if (sec != WIOS_ALL_SEC) {
		double sec_per_packet;
		sec_per_packet = ((priv->samp_au < 4) ? SEC_PER_PACKET1 : SEC_PER_PACKET2);
		sig->loops = (int)(sec/sec_per_packet);
		sig->loops++;
		capture_req_size = sig->loops*sig->buf_size*AD_CHANNELS/DA_CHANNELS;
	} else {
		sig->loops = sig->loops_org = priv->msize/sig->buf_size+1;
		capture_req_size = priv->msize*AD_CHANNELS/DA_CHANNELS;
	}

	/* 同期モードの時デバイスの初期化の際に必要 */
	ft->gain = set_gain_val;

	/* 録音フォーマットを揃える */
	common_samp_au = priv->samp_au;
	common_loops   = sig->loops;
	common_sec     = info->sec;
	
	return ft;

error:
	return NULL;
}

static wios_file_t * open_read(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	const char * ext = check_file_ext(path);

	if (ext == ".wav") return open_read_wav(path, dev, info);
	if (ext == ".och") return open_read_och(path, dev, info);
	if (ext == ".ich")
		WIOS_ERROR_S("DA file extension must be wav or och");

error:
	return NULL;
}

static wios_file_t * open_write(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info,
	const wios_fmtinfo_t   * fmt)
{
	int i;
	wios_file_t * ft;
	unsigned long	ich_header[4]={0x10000000,0,0,0};
	int rate = fmt->rate;
	int sec  = info->sec;
	int samp_au, loops;
	tdbd_priv_t * priv;
	wios_signal_t * sig;
	wios_execinfo_t * exec;

	if (check_output_file_name(path) == WIOS_EOF)
		goto error;

	/* samp_auを取得 */
	if (wios_mode == wios_rec || use_2dev_option_flag) {
		if ((samp_au = check_sampling_rate(rate)) == WIOS_EOF)
			WIOS_ERROR("not supported rate `%d'", rate);
	} else if (wios_mode == wios_sync && !use_2dev_option_flag) {
		if (check_sampling_rate(rate) == WIOS_EOF)
			WIOS_ERROR("not supported rate `%d'", rate);
		samp_au = common_samp_au;
	}

	/* loopsの値を指定 */
	if (sec != WIOS_ALL_SEC) {
		double sec_per_packet;
		for (i = 0; i < ARRAY_SIZE(tdbd_supported_rate); i++)
			if (rate == tdbd_supported_rate[i]) break;
		sec_per_packet = ((i < 4) ? SEC_PER_PACKET1 : SEC_PER_PACKET2);
		loops = (int)(sec/sec_per_packet);
		loops++;
	} else {
		if (wios_mode == wios_rec || use_2dev_option_flag)
			loops = NOT_USE;
		else 
			loops = common_loops;
	}

	ich_header[1] = samp_au;
	ich_header[2] = MODE_16CH;

	if ((ft = malloc(sizeof(wios_file_t))) == NULL)
		WIOS_ERROR_S("cannot allocate file structure");
	if (priv_init(ft) == WIOS_EOF)
		WIOS_ERROR_S("cannot allocate file structure");

	ft->io_type = wios_ad;
	ft->file_name = output_ich_file_name ? output_ich_file_name : TDBD_DEFAULT_ICHFILE_NAME;
	ft->dev_name  = dev  ? dev  : TDBD_DEFAULT_DEVICE_NAME;
	if (info) ft->signal.execinfo = *info;
	else wios_execinfo_init(&ft->signal.execinfo);
	
	priv = ft->priv;
	sig  = &ft->signal;
	exec = &ft->signal.execinfo;
	
	priv->samp_au = samp_au;
	sig->buf_size = audio_inpkt_tbl[samp_au]+4;
	sig->byte_per_item = byte_per_item[WIOS_PCM_FORMAT_S16];
	exec->periodsize = AD_BS_TO_PS(sig->buf_size);
	sig->loops = loops;

	ft->fmtinfo.rate = tdbd_supported_rate[priv->samp_au];
	ft->fmtinfo.channels = AD_CHANNELS;
	ft->fmtinfo.format = WIOS_PCM_FORMAT_S16;

	ft->gain = set_gain_val;

	/* 録音モード時に秒数が指定されていればcapture_req_sizeをここで設定 */
	/* 録音モードで秒数が指定されなかった場合はcapture_req_sizeは使わない */
	if ((wios_mode == wios_rec || use_2dev_option_flag) && sig->execinfo.sec != WIOS_ALL_SEC)
		capture_req_size = sig->byte_per_item*AD_CHANNELS*fmt->rate*sig->execinfo.sec;

	if (check_gain(set_gain_val) == WIOS_EOF)
		WIOS_ERROR("not supported gain `%d'", set_gain_val);

	if ((priv->buffer = priv->bp = (UCHAR *)malloc(sig->buf_size*32)) == NULL)
		WIOS_ERROR_S("read buffer allocate error");

	if (wios_mode == wios_sync && !use_2dev_option_flag && check_ad_rate(ft) != WIOS_SUCCESS)
		WIOS_ERROR_S("samplingrate must be same");

	if ((ft->fp = fopen(ft->file_name, "wb")) == NULL)
		WIOS_ERROR_S("ich file open error");

	if (fwrite(ich_header, sizeof(uchar), ICH_HEADER_SIZE, ft->fp) != ICH_HEADER_SIZE)
		WIOS_ERROR_S("ich file header write error");
	
	return ft;

error:
	return NULL;
}

static void * playback(void * arg)
{
	/* 再生準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t * ft = (wios_file_t *)arg;

	/* 再生準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start playback (%s)\n", ft->file_name);

	/* 再生ループ */
	tdbd_write(ft);

	/* 再生が完了したら録音用スレッドに知らせる */
	playback_end_flag = WIOS_TRUE;

	if (print_mesg_flag)
		fprintf(stdout, "End playback (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Playback_start_time:%lf\n", wios_get_time());

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	return NULL;
}

static void * capture(void * arg)
{
	/* 録音準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t * ft = (wios_file_t *)arg;

	/* 録音準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start capture (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Capture_start_Time:%lf\n", wios_get_time());

	/* 録音ループ */
	tdbd_read(ft);

	if (print_mesg_flag)
		fprintf(stdout, "End capture (%s)\n", ft->file_name);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	return NULL;
}

static void finish(wios_file_t * ft)
{
	tdbd_priv_t * priv = ft->priv;

	if (ft) {
		if (ft->io_type == wios_ad) {
			output_16ch_wav(ft);
		}
		if (ft->fp) {
			fclose(ft->fp);
			ft->fp = NULL;
		}
		if (ft->dp) {
			sinich_sys_close(ft->dp);
			ft->dp = NULL;
		}
		if (priv->buffer) {
			free(priv->buffer);
			priv->buffer = NULL;
		}
		if (ft->priv) {
			free(ft->priv);
			ft->priv = NULL;
		}

		free(ft);
		ft = NULL;
	}
}



/* STATIC FUNCTION */
/* -------------------------------------------------- */

static char * check_file_ext(const char * name)
{
	int len=strlen(name);
	char * ich_ext = ".ich";
	char * och_ext = ".och";
	char * wav_ext = ".wav";

	if (len > MAX_STRING_LENGTH)
		WIOS_ERROR_S("file name is too long");

	if (strstr(name, ich_ext) == name+len-4) return ich_ext;
	if (strstr(name, och_ext) == name+len-4) return och_ext;
	if (strstr(name, wav_ext) == name+len-4) return wav_ext;

	WIOS_ERROR_S("file extension must be ich or och or wav");

error:
	return NULL;
}

static int check_sampling_rate(int rate)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(tdbd_supported_rate); i++)
		if (tdbd_supported_rate[i] == rate) return i;

	return WIOS_EOF;
}

static int check_ad_rate(wios_file_t * ftad)
{
	if (ftda->fmtinfo.rate != ftad->fmtinfo.rate) return WIOS_EOF;
	else return WIOS_SUCCESS;
}

static int check_gain(int gain)
{
	if (gain >= GAIN_MIN && gain <= GAIN_MAX)
		return WIOS_SUCCESS;
	
	return WIOS_EOF;
}

static int check_output_file_name(const char * name)
{
	int len=strlen(name);
	int ich_flag = 0, wav_flag = 0;
	char * p;

	if ((p = check_file_ext(name)) == NULL) goto error;
	if (p == ".ich") ich_flag = 1;
	if (p == ".wav") wav_flag = 1;

	if (ich_flag) {
		strcpy(output_ich_file_name, name);
		strncpy(output_wav_file_name, name, len-3);
		output_wav_file_name[len-3] = 'w';
		output_wav_file_name[len-2] = 'a';
		output_wav_file_name[len-1] = 'v';
		output_wav_file_name[len] = '\0';
	} else if (wav_flag) {
		strncpy(output_ich_file_name, name, len-3);
		strcpy(output_wav_file_name, name);
		output_ich_file_name[len-3] = 'i';
		output_ich_file_name[len-2] = 'c';
		output_ich_file_name[len-1] = 'h';
		output_ich_file_name[len] = '\0';
	} else {
		strcpy(output_ich_file_name, name);
		strcpy(output_wav_file_name, name);
		output_wav_file_name[len] = '.';
		output_wav_file_name[len+1] = 'w';
		output_wav_file_name[len+2] = 'a';
		output_wav_file_name[len+3] = 'v';
		output_wav_file_name[len+4] = '\0';
		output_ich_file_name[len] = '.';
		output_ich_file_name[len+1] = 'i';
		output_ich_file_name[len+2] = 'c';
		output_ich_file_name[len+3] = 'h';
		output_ich_file_name[len+4] = '\0';
	}

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static int check_da_format(wios_file_t * ft) {
	wios_fmtinfo_t * fmt = &ft->fmtinfo;

	if (fmt->channels != DA_CHANNELS)
		WIOS_ERROR("DA channels `%d' must be %d", fmt->channels, DA_CHANNELS);
	if (check_sampling_rate(fmt->rate) == WIOS_EOF)
		WIOS_ERROR("not supported rate `%d'", fmt->rate);
	if (fmt->format != TDBD_FORAMT)
		WIOS_ERROR("DA format `%s' must be %s", wios_format_to_string(fmt->format), TDBD_FORAMT);

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static int priv_init(wios_file_t * ft)
{
	tdbd_priv_t * priv;

	if ((ft->priv = malloc(sizeof(tdbd_priv_t))) == NULL)
		goto error;

	priv = ft->priv;
	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

/* EP4 再生ループ */
static void tdbd_write(wios_file_t * ft)
{
	int audio_out=1, ret, n, i, cnt;
	int16_t * buf1;
	ushort buff_status;
	uchar ep4_buffspace_num=0;
	tdbd_priv_t * priv = ft->priv;
	wios_signal_t * sig = &ft->signal;

	if (wios_mode == wios_play || use_2dev_option_flag)
		sinich_sys_start(ft->dp);
	else if (wios_mode == wios_sync && !use_2dev_option_flag)
		sinich_all_start(ft->dp);

	while (audio_out) {
		if (loop_break_flag)
			break;

		if ((buff_status = sinich_sys_status(ft->dp)) == 0xFFFF)
			WIOS_ERROR_S("status read error");

		ep4_buffspace_num = (uchar)(buff_status>>8);

		if (ep4_buffspace_num) {
			if (audio_out) {
				if (ep4_buffspace_num > sig->loops)
					ep4_buffspace_num = sig->loops;

				/* n = min(sig->buf_size*ep4_buffspace_num, priv->msize-(priv->bp-priv->buffer));
				 * のコードだと再生が終了しなくなることがあるため(原因不明)、
				 * サンプルプログラムにならい以下のコードにしてある。 */
				n = sig->buf_size*ep4_buffspace_num;

				/* n[byte]のデータをデバイスに書き込む */
				ret = sinich_pcm_write(ft->dp, priv->bp, n);

				if (ret < 0) {
					DEVICE_RUN_WARN("write device error. missing data: unknown", NULL);
				} else if (ret != n) {
					DEVICE_RUN_WARN("device short write, write %d / %d bytes. missing data: %d bytes",
						ret, n, n-ret);
					ft->missing_byte_cnt += n-ret;
				}

				buf1 = (int16_t *)priv->bp;
				for (i = 0; i < ret/2; i++) {
					DA_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
					WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
				}
				ft->success_byte_cnt += min(ret, priv->msize-(priv->bp-priv->buffer));

				priv->bp += min(ret, priv->msize-(priv->bp-priv->buffer));
				sig->loops -= (ret/sig->buf_size);

				if (sig->loops <= 0 || priv->bp - priv->buffer >= priv->msize)
					audio_out = 0;
			}
		}
	}
	
	/* 再生が完了するまで待つ */
	while (1) {
		if ((buff_status = sinich_sys_status(ft->dp)) == 0xFFFF)
			WIOS_ERROR_S("status read error");

		ep4_buffspace_num = (uchar)(buff_status>>8);

		if (ep4_buffspace_num == 32) break;
	}

error:
	;
}

/* EP2 録音ループ */
static void tdbd_read(wios_file_t * ft)
{
	int audio_in=1, ret=0, n, i, cnt;
	int16_t * buf1;
	long long done = 0;
	ushort buff_status=0;
	uchar ep2_bufffill_num=0;
	tdbd_priv_t * priv = ft->priv;
	wios_signal_t * sig = &ft->signal;

	if (wios_mode == wios_rec || use_2dev_option_flag)
		sinich_sys_start(ft->dp);
	
	while (audio_in) {
		if (loop_break_flag)
			break;

		if ((buff_status = sinich_sys_status(ft->dp)) == 0xFFFF)
			WIOS_ERROR_S("status read error");

		ep2_bufffill_num = (uchar)buff_status;

		if (ep2_bufffill_num) {
			if (audio_in) {
				if (ep2_bufffill_num > sig->loops)
					ep2_bufffill_num = sig->loops;
				n = sig->buf_size*ep2_bufffill_num;

				/* n[byte]のデータをデバイスから読み込む */
				ret = sinich_pcm_read(ft->dp, priv->bp, n);

				if (ret < 0) {
					DEVICE_RUN_WARN("read device error. missing data: unknown", NULL);
				} else if (ret != n) {
					DEVICE_RUN_WARN("device short read, read %d / %d bytes. missing data: %d bytes",
						ret, n, n-ret);
					ft->missing_byte_cnt += n-ret;
				}

				/* 録音フレーム数を揃える */
				if (wios_mode != wios_rec || sig->execinfo.sec != WIOS_ALL_SEC) {
					if (ret > capture_req_size-done) {
						ret = capture_req_size-done;
						audio_in = 0;
					}
				}

				buf1 = (int16_t *)priv->bp;
				for (i = 0; i < ret/2; i++) {
					AD_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
					WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
				}
				ft->success_byte_cnt += ret;
				done += ret;

				if (fwrite((const void *)priv->bp, sizeof(uchar), ret, ft->fp) != (unsigned int)ret) {
					WIOS_ERROR_S("ich file write error");
					break;
				}
				sig->loops -= (ret/sig->buf_size);

				if (wios_mode != wios_rec || sig->execinfo.sec != WIOS_ALL_SEC) {
					if (sig->loops <= 0)
						audio_in = 0;
				}
			}
		}
	}

error:
	;
}

/* 録音したichファイルから16chWAVを出力する */
void output_16ch_wav(wios_file_t * ft)
{
	FILE * ichp;
	int nsamp = ft->signal.execinfo.periodsize/2;
	int16_t buf[AD_CHANNELS][MAX_FRAMES_PER_PACKET];
	int32_t fbuf[AD_CHANNELS*MAX_FRAMES_PER_PACKET];
	int i, j;
	int done = 0;
	int capture_req_nframes = capture_req_size/(ft->signal.byte_per_item*ft->fmtinfo.channels);
	int fpos_start, fpos_end, pcm_size;

	/* ichファイルを一旦閉じる */
	if (ft->fp) {
		fclose(ft->fp);
		ft->fp = NULL;
	}

	/* ichファイルを再び開く */
	if ((ichp = fopen(output_ich_file_name, "rb")) == NULL)
		WIOS_ERROR_S("ich file open error");
	
	/* ichファイルサイズを取得してサイズチェック */
	fseek(ichp, 0, SEEK_SET);
	fpos_start = ftell(ichp);
	fseek(ichp, 0, SEEK_END);
	fpos_end = ftell(ichp);
	pcm_size = fpos_end-fpos_start-ICH_HEADER_SIZE;

	if (pcm_size < capture_req_size)
		WIOS_ERROR_S("ich file size error");

	/* 録音モードで秒数が指定されなかったときはcapture_req_nframesを再設定 */
	if (wios_mode == wios_rec && ft->signal.execinfo.sec == WIOS_ALL_SEC) {
		capture_req_size = pcm_size;
		capture_req_nframes = capture_req_size/(ft->signal.byte_per_item*ft->fmtinfo.channels);
	}

	/* 出力WAVファイルを開く */
	ft->file_name = output_wav_file_name;
	if (lsf_open_write(ft) == NULL)
		goto error;

	/* ich => wav */
	fseek(ichp, 16, SEEK_SET);
	while (done < capture_req_nframes) {
		int t;
		fseek(ichp, 4, SEEK_CUR);

		for (t = 0; t < 2; t++) {
			/* 録音フレーム数と再生フレーム数を揃える */
			if (nsamp > capture_req_nframes-done)
				nsamp = capture_req_nframes-done;

			/* パケット抜けがあった場合は0データを入れる */
			memset(buf, 0, sizeof(buf));
			for (i = 0; i < AD_CHANNELS; i++)
				fread(buf[i], sizeof(int16_t), nsamp, ichp);

			for (i = 0; i < nsamp; i++)
				for (j = 0; j < AD_CHANNELS; j++)
					fbuf[i*AD_CHANNELS+j] = WIOS_CONVERT_16_TO_32BIT(buf[j][i]);

			lsf_file_write(ft, fbuf, nsamp);
			done += nsamp;
		}
	}

	/* 出力WAVファイルを閉じる */
	sf_close(ft->fp);
	ft->fp = NULL;

error:
	;
}

/*----------------------------------------------------*/
// Display connect audio board                        //
/*----------------------------------------------------*/
static int codec_control_set(uchar msg_flag, struct ich_io_t *io,int set_gain_val)
{
	int ret;
	struct codec_param_t	codec[4];
	uchar fpga_rdata;

	if((ret	= sinich_sys_act_info(io ,0 ,&fpga_rdata)) < 0){
		// printf("===== Audio board nothing! (ret(%d)<0) ===== \n", ret);
	}else{
		if( (unsigned char)fpga_rdata == audio_board_num_table[0] ){
			// printf("===== Audio board act! ===== \n");
			// Power down on
			codec[0].act_flag		= 0x0002;	// Bit1 set
			codec[0].adc_power_down	= 0;
			if((ret	= sinich_pcm_set_codec(io ,0 ,&codec[0])) < 0 ){
				printf("Codec set error!\n");
				return(-1);
			}
			Sleep(10);	// 10ms sleep
			// Power down off
			codec[0].act_flag		= 0x0002;	// Bit1 set
			codec[0].adc_power_down	= 1;
			if((ret	= sinich_pcm_set_codec(io ,0 ,&codec[0])) < 0 ){
				printf("Codec set error!\n");
				return(-1);
			}

			// Wait for ADC initialize end(Addtional function 2006/06/23)
			codec[0].act_flag = 0x0200;
			while(1){
				ret	= sinich_pcm_get_codec(io , 0 ,codec);
				if(ret < 0) {
					printf("Codec initialize status read error!!\n");
					return(-1);
				}
				if(((codec[0].operation_mode) & 0x08) != 0){
					break;
				}
			}

			if(set_gain_val != 0){
				// Set Gain(IPGA/ALC Mode Control 2)
				Sleep(100);	// 100ms sleep
				codec[0].act_flag		   = 0x0500;	// Set gain bit
				codec[0].input_pga_control = set_gain_val; // IPGA data
				codec[0].alc_mode_control2 = set_gain_val;		   // ALC Mode Control2 data
				if((ret	= sinich_pcm_set_codec(io ,0 ,&codec[0])) < 0 ){
					printf("Codec set error!\n");
					return(-1);
				}
			}

			if (msg_flag)
			{// For debug
				int		cnt1;
				int		cnt2;
				codec[0].act_flag		= 0x1FFF;	// All read
				codec[1].act_flag		= 0x1FFF;	// All read
				codec[2].act_flag		= 0x1FFF;	// All read
				codec[3].act_flag		= 0x1FFF;	// All read
				ret	= sinich_pcm_get_codec(io , 0 ,codec);
				if( ret < 0 ){
					printf("Get codec info error!\n");
					return(-1);
				}else{
					puts("CODEC:");
					for( cnt1=0 ; cnt1<4 ; cnt1++ ){
						for( cnt2=0 ; cnt2<0x0C ; cnt2++ ){
							printf("Address 0x%02X : Data 0x%02X\n",
								cnt2,*(((unsigned char *)&codec[cnt1])+cnt2));
						}
					}
					puts("");
				}
			}// For debug
		}
	}

	return 0;
}

/*----------------------------------------------------*/
// Display each descriptors                           //
/*----------------------------------------------------*/
static int DisplayDescriptors(HANDLE hDev)
{
	USB_DEVICE_DESCRIPTOR dsc;
	int ret_val = -1;

	ret_val = get_device_descriptor(hDev, &dsc);
	if (ret_val == -1) return (-1);

	printf("USB_DEVICE_DESCRIPTORS:\n");
	printf("bLength            0x%02X\n", dsc.bLength);
	printf("bDescriptorType    0x%02X\n", dsc.bDescriptorType);
	printf("bcdUSB             0x%04X\n", dsc.bcdUSB);
	printf("bDeviceClass       0x%02X\n", dsc.bDeviceClass);
	printf("bDeviceSubClass    0x%02X\n", dsc.bDeviceSubClass);
	printf("bDeviceProtocol    0x%02X\n", dsc.bDeviceProtocol);
	printf("bMaxPacketSize0    0x%02X\n", dsc.bMaxPacketSize0);
	printf("idVendor           0x%04X\n", dsc.idVendor);
	printf("iProduct           0x%04X\n", dsc.iProduct);
	printf("bcdDevice          0x%04X\n", dsc.bcdDevice);
	printf("iManufacturer      0x%02X\n", dsc.iManufacturer);
	printf("iProduct           0x%02X\n", dsc.iProduct);
	printf("iSerialNumber      0x%02X\n", dsc.iSerialNumber);
	printf("bNumConfigurations 0x%02X\n", dsc.bNumConfigurations);
	printf("\n");

	return(0);
}

/*----------------------------------------------------*/
// Display current sampling rate & ch (Audio)         //
/*----------------------------------------------------*/
static int DisplayAudioSampleRate(uchar msg_flag, uchar fpga_rdata)
{
	int samplerate;
	char data_endian[64] = {0};

	// Sampling rate selection //
	if ((fpga_rdata&0x07) ==  AUDIO_48K){
		samplerate = 48000;
	} else if((fpga_rdata&0x07) ==  AUDIO_32K){
		samplerate = 32000;
	} else if((fpga_rdata&0x07) ==  AUDIO_16K){
		samplerate = 16000;
	} else if((fpga_rdata&0x07) ==  AUDIO_8K){
		samplerate = 8000;
	} else if((fpga_rdata&0x07) ==  AUDIO_44K){
		samplerate = 44100;
	} else if((fpga_rdata&0x07) ==  AUDIO_22K){
		samplerate = 22050;
	} else{ // fpga_rdata&0x07 ==  AUDIO_11K
		samplerate = 11025;
	}

	// Audio data endian selection //
	if((fpga_rdata&0x20) == 0){
		strcpy(data_endian,"Little Endian");
	} else {
		strcpy(data_endian,"Big Endian");
	}

	// printf("Data endian                : %s\n",data_endian);
	// printf("Audio sampling frequency   : %dHz\n",samplerate);

	return 0;
}

static int hex_check(char *p)
{
	if (('0' != p[0]) || !(('x' == p[1]) || 'X' == p[1])) {
		return -1;
	}
	
	return 0;
}

// Alphabet Hexadecimal number charactor to Integer
static int ahextoi(char hex)
{
	if ((hex >= '0') && (hex <= '9')) return (hex - '0');
	else if ((hex >= 'a') && (hex <= 'f')) return (hex + 10 - 'a');
	else if ((hex >= 'A') && (hex <= 'F')) return (hex + 10 - 'A');

	return -1;
}

// Alphabet Hexadecimal number charactor to Integer
static int ahextoi2(char *hex_char, int digit)
{
	int base = 1, var, ret = 0;

	digit++;
	while(digit != 1) {
		var = ahextoi(hex_char[digit--]);
		if ((var<0) || (var>15)) {
			return -1;
		}
		ret += var * base;
		base <<= 4;
	}

	return ret;
}



/* GLOBAL FUNCTION */
/* -------------------------------------------------- */

int tdbd_device_check(const char * dev)
{
	static ICH_IO_T * dp = NULL;
	int ret;

	ret = (sinich_sys_open(&dp, (char *)dev) == 0) ? WIOS_TRUE : WIOS_FALSE;
	if (dp) sinich_sys_close(dp);

	return ret;
}

wios_format_handler_t const * wios_set_tdbd_format_handler(void)
{
	static wios_format_handler_t const handler = {
		device_init,
		open_read, open_write,
		playback, capture,
		finish
	};

	return &handler;
}

/*---------------------------------------------------------------*/
/*----------------------*/
/*	Register read	 	*/
/*----------------------*/
int tdbd_register_read(char * reg_rd_addr_name)
{
	int com_reg_read = 1;
	int ret_val;
	int reg_rd_addr=0, reg_wr_addr=0;
	ICH_IO_T * ted_usb_sys;
	uchar fpga_rdata=0;

	//-- Driver open for control transfer(EP0) --//
	if((ret_val = sinich_sys_open(&ted_usb_sys, device_name)) != 0){
		printf("Driver open error!\n");
		return WIOS_EOF;
	}

	//-- Audio,I/O test is invalid --//
	/*audio_in = 0;
	audio_out= 0;*/

	//-------------------------------//
	ret_val = hex_check(reg_rd_addr_name);
	if (ret_val != 0) {com_reg_read = 0; goto END_REG_READ;}
	reg_rd_addr = ahextoi2(reg_rd_addr_name, 4);
	if (reg_rd_addr == -1) {com_reg_read = 0; goto END_REG_READ;}
	if (reg_rd_addr_name[3] == '8') {
		ret_val = sinich_sys_get_reg(ted_usb_sys, reg_rd_addr, &fpga_rdata);
		if (ret_val != 0) {com_reg_read = 0; goto END_REG_READ;}
	}else {
//-- 0x40xx - 0x47xx : support --//
#ifdef ALLREG_ACEESS
		ret_val = sinich_sys_get_allreg(ted_usb_sys, reg_rd_addr, &fpga_rdata);
		if (ret_val != 0) {com_reg_read = 0; goto END_REG_READ;}
		goto END_REG_READ;
#endif 


//-- 0x40xx - 0x47xx : no support --//
		printf("Error : Please input system board register(0x48xx)\n\n");
		/* goto END_PROC; */
		goto CLOSE_DEV;
	}
END_REG_READ:
	if (com_reg_read) printf("\nRegister read  OK     Address : 0x%04X, Data : 0x%02X\n\n", reg_rd_addr, fpga_rdata);
	else printf("\nRegister read  NG     Address : 0x%04X, Data : ----\n\n", reg_rd_addr);

CLOSE_DEV:
	if(ted_usb_sys!=NULL)		sinich_sys_close(ted_usb_sys);
	
	/* goto END_PROC; */
	return WIOS_EOF;
}

/*---------------------------------------------------------------*/
/*----------------------*/
/*	Register write		*/
/*----------------------*/
int tdbd_register_write(char * reg_wr_addr_name, char * reg_wr_data_name)
{
	//-- Audio,I/O test is invalid --//
	/*audio_in = 0;
	audio_out= 0;*/
	int com_reg_write = 1;
	int ret_val;
	int reg_rd_addr=0, reg_wr_addr=0;
	ICH_IO_T * ted_usb_sys;
	uchar fpga_rdata=0;

	//-- Driver open for control transfer(EP0) --//
	if((ret_val = sinich_sys_open(&ted_usb_sys, device_name)) != 0){
		printf("Driver open error!\n");
		return WIOS_EOF;
	}

//-------------------------------//
	ret_val = hex_check(reg_wr_addr_name);
	if (ret_val != 0) {com_reg_write = 0; goto END_REG_WRITE;}
	ret_val = hex_check(reg_wr_data_name);
	if (ret_val != 0) {com_reg_write = 0; goto END_REG_WRITE;}
	reg_wr_addr = ahextoi2(reg_wr_addr_name, 4);
	if (reg_wr_addr == -1) {com_reg_write = 0; goto END_REG_WRITE;}
	fpga_rdata = ahextoi2(reg_wr_data_name, 2);
	if (reg_wr_addr_name[3] == '8') {
		ret_val = sinich_sys_set_reg(ted_usb_sys, reg_wr_addr, fpga_rdata);
		if (ret_val != 0) {com_reg_write = 0; goto END_REG_WRITE;}
	}else {
//-- 0x40xx - 0x47xx : support --//
#ifdef ALLREG_ACEESS
		ret_val = sinich_sys_set_allreg(ted_usb_sys, reg_wr_addr, fpga_rdata);
		if (ret_val != 0) {com_reg_read = 0; goto END_REG_WRITE;}
		goto END_REG_WRITE;
#endif 

//-- 0x40xx - 0x47xx : no support --//
		printf("Error : Please input system board register(0x48xx)\n\n");
		/* goto END_PROC; */
		goto CLOSE_DEV;
	}
END_REG_WRITE:
	if (com_reg_write) printf("\nRegister write OK     Address : 0x%04X, Data : 0x%02X\n\n", reg_wr_addr, fpga_rdata);
	else printf("\nRegister write NG     Address : 0x%04X, Data : ----\n\n", reg_wr_addr);

CLOSE_DEV:
	if(ted_usb_sys!=NULL)		sinich_sys_close(ted_usb_sys);

	return WIOS_EOF;
}
